/*
 *    GeoTools - The Open Source Java GIS Toolkit
 *    http://geotools.org
 *
 *    (C) 2001-2008, Open Source Geospatial Foundation (OSGeo)
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation;
 *    version 2.1 of the License.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 */
package org.geotools.resources.i18n;

import java.util.Locale;
import java.util.MissingResourceException;
import org.geotools.resources.IndexedResourceBundle;

/**
 * Base class for locale-dependent resources. Instances of this class should never been created
 * directly. Use the factory method {@link #getResources} or use static convenience methods instead.
 *
 * @since 2.2
 * @source $URL$
 * @version $Id$
 * @author Martin Desruisseaux (IRD)
 */
public class Descriptions extends IndexedResourceBundle {
    /**
     * Returns resources in the given locale.
     *
     * @param locale The locale, or {@code null} for the default locale.
     * @return Resources in the given locale.
     * @throws MissingResourceException if resources can't be found.
     */
    public static Descriptions getResources(Locale locale) throws MissingResourceException {
        if (locale == null) {
            locale = Locale.getDefault();
        }
        return (Descriptions) getBundle(Descriptions.class.getName(), locale);
        /*
         * We rely on cache capability of ResourceBundle.
         */
    }

    /**
     * Gets a string for the given key from this resource bundle or one of its parents.
     *
     * @param key The key for the desired string.
     * @return The string for the given key.
     * @throws MissingResourceException If no object for the given key can be found.
     */
    public static String format(final int key) throws MissingResourceException {
        return getResources(null).getString(key);
    }

    /**
     * Gets a string for the given key are replace all occurence of "{0}" with values of {@code
     * arg0}.
     *
     * @param key The key for the desired string.
     * @param arg0 Value to substitute to "{0}".
     * @return The formatted string for the given key.
     * @throws MissingResourceException If no object for the given key can be found.
     */
    public static String format(final int key, final Object arg0) throws MissingResourceException {
        return getResources(null).getString(key, arg0);
    }

    /**
     * Gets a string for the given key are replace all occurence of "{0}", "{1}", with values of
     * {@code arg0}, {@code arg1}.
     *
     * @param key The key for the desired string.
     * @param arg0 Value to substitute to "{0}".
     * @param arg1 Value to substitute to "{1}".
     * @return The formatted string for the given key.
     * @throws MissingResourceException If no object for the given key can be found.
     */
    public static String format(final int key, final Object arg0, final Object arg1)
            throws MissingResourceException {
        return getResources(null).getString(key, arg0, arg1);
    }

    /**
     * Gets a string for the given key are replace all occurence of "{0}", "{1}", with values of
     * {@code arg0}, {@code arg1}, etc.
     *
     * @param key The key for the desired string.
     * @param arg0 Value to substitute to "{0}".
     * @param arg1 Value to substitute to "{1}".
     * @param arg2 Value to substitute to "{2}".
     * @return The formatted string for the given key.
     * @throws MissingResourceException If no object for the given key can be found.
     */
    public static String format(
            final int key, final Object arg0, final Object arg1, final Object arg2)
            throws MissingResourceException {
        return getResources(null).getString(key, arg0, arg1, arg2);
    }
}
